/****************************************************************************
** ui.h extension file, included from the uic-generated form implementation.
**
** If you want to add, delete, or rename functions or slots, use
** Qt Designer to update this file, preserving your code.
**
** You should not define a constructor or destructor in this file.
** Instead, write your code in functions called init() and destroy().
** These will automatically be called by the form's constructor and
** destructor.
*****************************************************************************/
#include "qurl.h"
#include "qsettings.h"
extern "C" {
  #include "gpsdclient.h"
}

void MainForm::writeSettings()
 {
     QSettings settings;
     settings.setPath("GPSD", "qtgpsc");
     settings.writeEntry("/gpsd/qtgpsc/units/position", unitsPosition->name());
     settings.writeEntry("/gpsd/qtgpsc/units/length", unitsLength->name());
     settings.writeEntry("/gpsd/qtgpsc/units/speed", unitsSpeed->name());
 }

void MainForm::readSettings()
 {
    // There *has* got to be a better way of doing this! :(
    QString val;
    QSettings settings;
    settings.setPath("GPSD", "qtgpsc");
    val = settings.readEntry("/gpsd/qtgpsc/units/position");
    if (val == "unitsPositionDegrees")
        unitsPositionDegrees->setOn(true);
    else if (val == "unitsPositionDMS")
        unitsPositionDMS->setOn(true);
    else if (val == "unitsPositionDM")
        unitsPositionDM->setOn(true);
    else if (val == "unitsPositionRadians")
        unitsPositionRadians->setOn(true);
    
    val = settings.readEntry("/gpsd/qtgpsc/units/length");
    if (val == "unitsLengthImperial")
        unitsLengthImperial->setOn(true);
    else if (val == "unitsLengthMetric")
        unitsLengthMetric->setOn(true);
    
    val = settings.readEntry("/gpsd/qtgpsc/units/speed");
    if (val == "unitsSpeedMetresSecond")
        unitsSpeedMetresSecond->setOn(true);
    else if (val == "unitsSpeedKilometresHour")
        unitsSpeedKilometresHour->setOn(true);
    else if (val == "unitsSpeedMilesHour")
        unitsSpeedMilesHour->setOn(true);
    else if (val == "unitsSpeedKnots")
        unitsSpeedKnots->setOn(true);

    if (val = getenv("GPSD_UNITS")) {
        if (val == "imperial") {
            unitsLengthImperial->setOn(true);
            unitsSpeedMilesHour->setOn(true);
        } else if (val == "nautical") {
            unitsLengthImperial->setOn(true);
            unitsSpeedKnots->setOn(true);
        } else if (val == "metric") {
            unitsLengthMetric->setOn(true);
            unitsSpeedKilometresHour->setOn(true);
        }
    }
 }


void MainForm::closeEvent(QCloseEvent *)
{
    fileExit();
}

void MainForm::fileExit()
{
    gpsrec.gps_disconnect();
    writeSettings();
    QApplication::exit( 0 );
}

void MainForm::helpAbout()
{
    QMessageBox::about(this, "About qtGPSc", QString(
            "<p>Client for the gpsd GPS server<br/>\n"
            "&lt;http://gpsd.berlios.de/&gt;<br/>\n"
            " based on the Qt toolkit.</p>\n"
            "<p>Version %1</p>\n"
            "<p>&copy;2007 Diego Berge &lt;<a href='mailto:gpsd@nippur.net'>gpsd@nippur.net</a>&gt;</p>").arg(qtgpscVersion));
}

// Initialise form
void MainForm::init()
{
    qtgpscVersion = "0.2.2";
    clipboard = QApplication::clipboard();
    if ( clipboard->supportsSelection() )
        clipboard->setSelectionMode( TRUE );
    isConnected = false;
    // Set default units
    // TODO - Honour $GPSD_UNITS
    unitsPosition = unitsPositionDMS;
    unitsLength = unitsLengthMetric;
    unitsSpeed = unitsSpeedKilometresHour;
    // and then try and load saved prefs
    readSettings();
    // qtgpsc accepts one command line argument of the form
    // "gpsd://host:port/device" or "/device"
    if (qApp->argc() == 2) {
        QUrl url(qApp->argv()[1]);
        if (url.isValid() && (url.protocol() == "gpsd" || url.protocol() == "file")) {
            if (url.hasHost())
                host_edit->setText(url.host());
            if (url.hasPort())
                port_edit->setText(QString::number(url.port()));
            gpsdConnect();
            // A device can be selected only once a connection has been
            // established.
            if (url.hasPath() && url.path() != "/" && gpsrec.isConnected())
                gpsrec.setCurrentDevice(url.path());
        }
    }
}

// Connect or disconnect from server
void MainForm::gpsdConnect()
{
    if (gpsrec.isConnected()) {
        // Disconnect
        isConnected = false;
        host_edit->setEnabled(true);
        port_edit->setEnabled(true);
        editCopyPositionAction->setEnabled(false);
        editCopyAsKMLAction->setEnabled(false);
        gpsrec.gps_disconnect();
        rawDataSocket.close();
        connect_btn->setOn(false);
        connect_btn->setIconSet(QIconSet( QPixmap::fromMimeSource( "connect_no.png" ) ));
        QToolTip::add(connect_btn, "Connect to server (Ctrl-T)");
        dev_combo->clear();
        rawDataCombo->setCurrentItem(0);
        setCaption("qtGPSc");
        statusBar()->message("Disconnected");
    } else {
        // Attempt to connect
        if (gpsrec.gps_connect(host_edit->text(), port_edit->text())) {
            statusBar()->message("Connecting...");
            isConnected = true;
            connect_btn->setIconSet(QIconSet( QPixmap::fromMimeSource( "connect_established.png" ) ));
            QToolTip::add(connect_btn, "Disconnect from server (Ctrl-T)");
            host_edit->setEnabled(false);
            port_edit->setEnabled(false);
            editCopyPositionAction->setEnabled(true);
            editCopyAsKMLAction->setEnabled(true);
            setCaption(QString("qtGPSc - %1:%2").arg(host_edit->text()).arg(port_edit->text()));
            connect_btn->setOn(true);
            setJitterMode(noJitterMode->isOn());
            statusBar()->message("Connected");
            rawdata->clear();
            // Populate device list
            if (*(gpsrec.getGpsData()->gps_device)) {
                dev_combo->insertItem(QString(gpsrec.getGpsData()->gps_device));
                setCaption(caption().append(dev_combo->currentText()));
            }
            connect(&gpsrec, SIGNAL(dataReceived()), this, SLOT(updateViews()));
            connect(&gpsrec, SIGNAL(deviceListChanged()), this, SLOT(refreshDevices()));
            connect(&gpsrec, SIGNAL(deviceChanged()), this, SLOT(refreshDevices()));
            connect(&rawDataSocket, SIGNAL(readyRead()), this, SLOT(readRawData()));
        } else {
            // Connection unsuccessful
            isConnected = false;
            connect_btn->setOn(false);
            statusBar()->message("Unable to connect to gpsd server");
        }
    }
}

// Raw data display.  This opens a separate connection to the server
// and issues the appropriate "r" command for the type of capture requested
void MainForm::dataCapture(int captureType)
{
    rawDataSocket.close();
    switch (captureType) {
        case 0: // No capture
            break;
        case 1: // Capture GPSD
            rawDataSocket.connectToHost(host_edit->text(), port_edit->text().toUInt());
            rawDataSocket.writeBlock("idfbw+\n", 7);
            break;
        case 2: // Capture NMEA
            rawDataSocket.connectToHost(host_edit->text(), port_edit->text().toUInt());
            rawDataSocket.writeBlock("r+\n", 3);
            break;
        case 3: // Capture Raw -- not implemented
        case 4:
            rawDataSocket.connectToHost(host_edit->text(), port_edit->text().toUInt());
            rawDataSocket.writeBlock("r2\n", 3);
        default:
            break;
    }
}

// Raw data available, dump to raw data pane.
void MainForm::readRawData()
{
    QCString s;
    s.resize(rawDataSocket.bytesAvailable()+1);
    rawDataSocket.readBlock(s.data(), rawDataSocket.bytesAvailable());
    if (rawDataCombo->currentItem() != 4)
        rawdata->append(s);
    else {
        for(uint n=0; n<s.size(); n++) {
            if (n%20 == 0)
                rawdata->append(QString("\n"));
            rawdata->append(QString().sprintf(" %02x ", s[n]));
        }
    }
}
        
// Update main pane, call updateSkyView() at the end of it
void MainForm::updateViews()
{
    char s[128];
    QString text;
    
    gps_data = gpsrec.getGpsData();
    
    if (gps_data == 0) {
        return;
    }
    
    // Update lat/lon
    if (unitsPosition == unitsPositionDMS) {
        latitudeEntry->setText(
                QString(deg_to_str(deg_ddmmss, fabs(gps_data->fix.latitude)))
                .append(gps_data->fix.latitude < 0 ? 'S' : 'N'));
        longitudeEntry->setText(
                QString(deg_to_str(deg_ddmmss, fabs(gps_data->fix.longitude)))
                .append(gps_data->fix.longitude < 0 ? 'W' : 'E'));
    } else if (unitsPosition == unitsPositionDM) {	
        latitudeEntry->setText(
                QString("%1 %2' %3")
                .arg(static_cast<int>(trunc(fabs(gps_data->fix.latitude))))
                .arg(fmod(fabs(gps_data->fix.latitude),1)*60, 0, 'f', 5)
                .arg(gps_data->fix.latitude < 0 ? 'S' : 'N'));
        longitudeEntry->setText(
                QString("%1 %2' %3")
                .arg(static_cast<int>(trunc(fabs(gps_data->fix.longitude))))
                .arg(fmod(fabs(gps_data->fix.longitude),1)*60, 0, 'f', 5)
                .arg(gps_data->fix.longitude < 0 ? 'S' : 'N'));
    } else if (unitsPosition == unitsPositionDegrees) {
        latitudeEntry->setText(QString("%1").arg(gps_data->fix.latitude, 0, 'f', 7));
        longitudeEntry->setText(QString("%1").arg(gps_data->fix.longitude, 0, 'f', 7));
    } else if (unitsPosition == unitsPositionRadians) {
        latitudeEntry->setText(QString("%1").arg(gps_data->fix.latitude*M_PI/180, 0, 'f', 9));
        longitudeEntry->setText(QString("%1").arg(gps_data->fix.longitude*M_PI/180, 0, 'f', 9));
    } else {
        latitudeEntry->setText(QString("?"));
        longitudeEntry->setText(QString("?"));
    }
    
    // Update Height
    if (unitsLength == unitsLengthMetric) {
        heightEntry->setText(QString("%1m").arg(gps_data->fix.altitude, 0, 'f', 3));
    } else if (unitsLength == unitsLengthImperial) {
        heightEntry->setText(QString("%1ft")
                             .arg(gps_data->fix.altitude*METERS_TO_FEET, 0, 'f', 1));
    } else {
        heightEntry->setText(QString("?"));
    }
    
    // Update Speed and Bearing
    if (gps_data->fix.speed > 0.1) {
        if (unitsSpeed == unitsSpeedMetresSecond) {
            speedEntry->setText(QString("%1m/s").arg(gps_data->fix.speed, 0, 'f', 2));
        } else if (unitsSpeed == unitsSpeedKilometresHour) {
            speedEntry->setText(QString("%1km/h").arg(gps_data->fix.speed*MPS_TO_KPH, 0, 'f', 0));
        } else if (unitsSpeed == unitsSpeedMilesHour) {
            speedEntry->setText(QString("%1mph").arg(gps_data->fix.speed*MPS_TO_MPH, 0, 'f', 0));
        } else if (unitsSpeed == unitsSpeedKnots) {
            speedEntry->setText(QString("%1kt").arg(gps_data->fix.speed*MPS_TO_KNOTS, 0, 'f', 0));
        } else {
            speedEntry->setText(QString("?"));
        }
        
        if (unitsPosition == unitsPositionDMS) {
            bearingEntry->setText(
                    QString(deg_to_str(deg_ddmmss, gps_data->fix.track)));
        } else if (unitsPosition == unitsPositionDM) {	
            bearingEntry->setText(
                    QString("%1 %2'")
                    .arg(static_cast<int>(trunc(fabs(gps_data->fix.track))))
                    .arg(fmod(fabs(gps_data->fix.track),1)*60, 0, 'f', 5));
        } else if (unitsPosition == unitsPositionDegrees) {
            bearingEntry->setText(QString("%1").arg(gps_data->fix.track, 0, 'f', 7));
        } else if (unitsPosition == unitsPositionRadians) {
            bearingEntry->setText(QString("%1").arg(gps_data->fix.track*M_PI/180, 0, 'f', 9));
        } else {
            bearingEntry->setText(QString("?"));
        }
    } else {
        speedEntry->clear();
        bearingEntry->clear();
    }
    
     // Update Time
    timeEntry->setText(QString(gpsrec.fixISOTime()).append("Z"));
    
    // Update DOPs
    if (isnan(gps_data->pdop)==0) {
        snprintf(s, sizeof(s), "P%.1f H%.1f V%.1f T%.1f G%.1f",
                gps_data->pdop, gps_data->hdop, gps_data->vdop, gps_data->tdop, gps_data->gdop);
        dopEntry->setText(QString(s));
    } else {
        dopEntry->setText(QString("?"));
    }
    
    // Update RMS
    {
        // Set conversion factor for RMS values
        double f;
        int p;
        char *s;
        if (unitsLength == unitsLengthMetric) {
            f = 1; s = "m"; p = 3;
        } else if (unitsLength == unitsLengthImperial) {
            f = METERS_TO_FEET; s = "ft"; p = 1;
        } else {
            f = 0; s = ""; p = 0;
        }
        
        if (isnan(gps_data->fix.eph) || f == 0) {
            rmsEntry->setText(QString("H?"));
        } else {
            rmsEntry->setText(QString("H%1%2").arg(gps_data->fix.eph*f, 0, 'f', p).arg(s));
        }
        if (isnan(gps_data->fix.epv) || f == 0) {
            rmsEntry->insert(" V?");
        } else {
            rmsEntry->insert(QString(" V%1%2").arg(gps_data->fix.eph*f, 0, 'f', p).arg(s));
        }
    }
    
    // Update sat table
    if (satTable->numRows() != gps_data->satellites)
        satTable->setNumRows(gps_data->satellites);
    for (int n=0; n<gps_data->satellites; n++) {
        satTable->setText(n, 0, QString("%1").arg(gps_data->PRN[n], 3));
        satTable->setText(n, 1, QString("%1").arg(gps_data->azimuth[n], 3));
        satTable->setText(n, 2, QString("%1").arg(gps_data->elevation[n], 2));
        satTable->setText(n, 3, QString("%1").arg(gps_data->ss[n], 2));
    }
    
    // Update skyview
    updateSkyview(skyviewFrame);
}

// Update the satellite polar graph
void MainForm::updateSkyview(QWidget *w)
{
    uint margin = 4;
    uint spacing = 2;
    uint radius = (std::min(w->width(), w->height())-margin*2-spacing*2)/2;
    uint centre = radius+margin;
    QPixmap p((radius+margin)*2, (radius+margin)*2);
    QPainter paint(&p);

    
    // Start with a clean slate
    p.fill(w->paletteBackgroundColor());
    
    // Draw background
    paint.setPen(QPen(Qt::darkGray, 2));
    paint.drawEllipse(margin, margin, radius*2, radius*2);
    paint.setPen(QPen(Qt::gray, 1, Qt::DotLine));
    paint.drawEllipse(static_cast<int>(centre-radius*(1.0/3)),
                      static_cast<int>(centre-radius*(1.0/3)),
                      static_cast<int>(radius*(2.0/3)),
                      static_cast<int>(radius*(2.0/3)));
    paint.drawEllipse(static_cast<int>(centre-radius*(2.0/3)),
                      static_cast<int>(centre-radius*(2.0/3)),
                      static_cast<int>(radius*(4.0/3)),
                      static_cast<int>(radius*(4.0/3)));
    for (double n=0; n<M_PI; n+=M_PI/4) {
        paint.drawLine(static_cast<int>(centre+radius*cos(n)),
                       static_cast<int>(centre+radius*sin(n)),
                       static_cast<int>(centre-radius*cos(n)),
                       static_cast<int>(centre-radius*sin(n)));
    }
    
    // Draw satellites
    if (gps_data) { // We might not be connected to a GPS
        QFont font = paint.font();
        font.setPointSize(10);
        paint.setFont(font);
        for(int i=0; i<gps_data->satellites; i++) {
            double x, y;
            x = centre 
                + cos((gps_data->azimuth[i]-90)*DEG_2_RAD)*(90-gps_data->elevation[i])*radius/90.0;
            y = centre
                + sin((gps_data->azimuth[i]-90)*DEG_2_RAD)*(90-gps_data->elevation[i])*radius/90.0;
            if (gps_data->PRN[i] > GPS_PRNMAX) // Not a GPS satellite
                paint.setPen(QPen(Qt::blue, 4));
            else if (gps_data->used[i])
                paint.setPen(QPen(Qt::red, 3));
            else
                paint.setPen(QPen(Qt::darkGray, 3));
            paint.drawEllipse(static_cast<int>(x-1),static_cast<int>(y-1),2,2);
            paint.setPen(QPen(Qt::black));
            paint.drawText(static_cast<int>(x+4), static_cast<int>(y+4),
                           QString("%1").arg(gps_data->PRN[i]));
        }
    }
    
    w->setPaletteBackgroundPixmap(p);
}

// Callback for Units -> Position
void MainForm::setUnitsPosition( QAction *a )
{
    unitsPosition = a;
}

// Callback for Units -> Length
void MainForm::setUnitsLength( QAction *a )
{
    unitsLength = a;
}

// Callback for Units -> Speed
void MainForm::setUnitsSpeed( QAction *a )
{
    unitsSpeed = a;
}

// Refresh the device list
void MainForm::refreshDevices()
{
    gps_data = gpsrec.getGpsData();
    dev_combo->clear();
    dev_combo->insertStrList(gpsrec.getDeviceList());
    setCaption(QString("qtGPSc - %1:%2%3")
               .arg(host_edit->text())
               .arg(port_edit->text())
               .arg(dev_combo->currentText()));
    if (gps_data->gps_id != 0 && gps_data->gps_id[0] != '\0') {
        statusBar()->message(QString("Connected to %1").arg(gps_data->gps_id));
    } else {
        statusBar()->message(QString("Connected"));
    }
}

// Callback for File -> Jitter Mode
void MainForm::setJitterMode( bool s )
{
    gpsrec.setJitter(s);
}

// Callback for Edit -> Copy Position
void MainForm::editCopyPosition()
{
    gps_data = gpsrec.getGpsData();
    if (gps_data == NULL)
        return;
    clipboard->setText(QString("%1, %2").arg(gps_data->fix.latitude).arg(gps_data->fix.longitude), QClipboard::Clipboard);
}

// Callback for Edit -> Copy As KML
void MainForm::editCopyAsKML()
{
    gps_data = gpsrec.getGpsData();
    if (gps_data == NULL)
        return;
    QString kml1 = "<?xml version=\"1.0\" encoding=\"UTF-8\"?>\n"
"<kml xmlns=\"http://earth.google.com/kml/2.2\">\n"
"<Document>\n"
"	<name>GPSD Position</name>\n"
"	<StyleMap id=\"msn_star\">\n"
"		<Pair>\n"
"			<key>normal</key>\n"
"			<styleUrl>#sn_star</styleUrl>\n"
"		</Pair>\n"
"		<Pair>\n"
"			<key>highlight</key>\n"
"			<styleUrl>#sh_star</styleUrl>\n"
"		</Pair>\n"
"	</StyleMap>\n"
"	<Style id=\"sn_star\">\n"
"		<IconStyle>\n"
"			<color>ff7faaff</color>\n"
"			<Icon>\n"
"				<href>http://maps.google.com/mapfiles/kml/shapes/star.png</href>\n"
"			</Icon>\n"
"		</IconStyle>\n"
"		<ListStyle>\n"
"		</ListStyle>\n"
"	</Style>\n"
"	<Style id=\"sh_star\">\n"
"		<IconStyle>\n"
"			<color>ff7faaff</color>\n"
"			<scale>1.18182</scale>\n"
"			<Icon>\n"
"				<href>http://maps.google.com/mapfiles/kml/shapes/star.png</href>\n"
"			</Icon>\n"
"		</IconStyle>\n"
"		<ListStyle>\n"
"		</ListStyle>\n"
"	</Style>\n"
"	<Placemark>\n"
"		<name>%1</name>\n"
"		<Snippet>%2</Snippet>\n"
"		<description><![CDATA[\n"
"<table>\n"
"<tr><th>Latitude</th><td>%3</td></tr>\n"
"<tr><th>Longitude</th><td>%4</td></tr>\n"
"<tr><th>Altitude</th><td>%5</td></tr>\n"
"<tr><th>Time</th><td>%6</td></tr>\n"
"<tr><th>Speed</th><td>%7</td></tr>\n"
"<tr><th>Bearing</th><td>%8</td></tr>\n"
"<tr><th>DOP</th><td>%9</td></tr>\n";
    QString kml2 = "<tr><th>RMS</th><td>%1</td></tr>\n"
"</table>\n"
"]]></description>\n"
"		<styleUrl>#msn_star</styleUrl>\n"
"		<Point>\n"
"			<coordinates>%2,%3,%4</coordinates>\n"
"		</Point>\n"
"	</Placemark>\n"
"</Document>\n"
"</kml>\n";

    QString name = QString("%1:%2%3").arg(host_edit->text())
        .arg(port_edit->text())
        .arg(dev_combo->currentText());

    QString kml = kml1.arg(name)
        .arg(timeEntry->text())
        .arg(latitudeEntry->text())
        .arg(longitudeEntry->text())
        .arg(heightEntry->text())
        .arg(timeEntry->text())
        .arg(speedEntry->text())
        .arg(bearingEntry->text())
        .arg(dopEntry->text());
    kml += kml2.arg(rmsEntry->text())
        .arg(gps_data->fix.longitude)
        .arg(gps_data->fix.latitude)
        .arg(gps_data->fix.altitude);
    
    clipboard->setText(kml, QClipboard::Clipboard);
}

// Callback for device combo selection change
void MainForm::deviceChanged( int n )
{
    gpsrec.setCurrentDevice(dev_combo->text(n));
}
