#include <iostream>
#include <string.h>
#include "qtgpsmm.h"

static qtGpsd *qtGpsd_instance = 0;

qtGpsd::qtGpsd()
{
    qtGpsd_instance = this;
    gps_data = 0;
}

bool qtGpsd::gps_connect(const char *host, const char *port) {
    gps_data = gps_open(host,port);
    if (gps_data == 0)
        return 0;
    
    gps_query(gps_data, "w+xqk?fj0");
    readNotify = new QSocketNotifier(gps_data->gps_fd, QSocketNotifier::Read);
    readNotify->setEnabled(true);
    connect(readNotify, SIGNAL( activated(int) ), this, SLOT( handleInput(int) ) );
    return 1;
}

void qtGpsd::gps_disconnect() {
    if (gps_data != 0) {
        readNotify->disconnect();
        delete readNotify;
        gps_close(gps_data);
        gps_data = 0;
    }
}

void qtGpsd::callback(struct gps_data_t* p, char* buf, size_t len, int level)
{
    if (qtGpsd_instance == 0)
        return;
    if (p==0) {
        // Error polling gpsd
        qtGpsd_instance->buffer = QString("***");
    } else {
        qtGpsd_instance->gps_data = p;
        qtGpsd_instance->buffer = QString::fromAscii(buf, len);
        qtGpsd_instance->level = level;
    }
    emit qtGpsd_instance->dataReceived();
}

void qtGpsd::handleInput(int socket)
{
    struct gps_data_t old_gps_data = *gps_data;
    if (gps_poll(gps_data) < 0) {
        // alert(ggpsc_main, "gGPS Connection Error", "\nRead error on server socket.\n");
        return;
    }
    gps_query(gps_data, "qs"); /* DOPs and status */
    
    if (old_gps_data.ndevices != gps_data->ndevices) {
        emit deviceListChanged();
    }
    
    if (strcmp(old_gps_data.gps_device, gps_data->gps_device) != 0) {
        emit deviceChanged();
    }
    
    if (strcmp(old_gps_data.gps_id, gps_data->gps_id) != 0) {
        emit deviceChanged();
    }
    
    emit dataReceived();
    return;
}

QStrList qtGpsd::getDeviceList()
{
    QStrList gpsDevices;
    for(int n=0; n<gps_data->ndevices; n++) {
        gpsDevices.append(gps_data->devicelist[n]);
    }
    return gpsDevices;
}

char* qtGpsd::getCurrentDevice()
{
    if (gps_data != 0) {
        return gps_data->gps_device;
    } else {
        return 0;
    }
}

void qtGpsd::setCurrentDevice(const char* device)
{
    if (gps_data != 0) {
        gps_query(gps_data, QString("f=%1").arg(device).ascii());
    }
}

QString qtGpsd::fixISOTime() {
    QDateTime dt;
    if (gps_data) {
        dt.setTime_t(static_cast<uint>(gps_data->fix.time), Qt::UTC);
        return dt.toString(Qt::ISODate);
    } else {
        return QString();
    }
}

void qtGpsd::setJitter(bool s)
{
    if (gps_data) {
        gps_query(gps_data, s?"j1":"j0");
    }
}
